from depyf import decompile, Decompiler
import unittest
import dis
import sys
from dataclasses import dataclass
from copy import deepcopy

# def test_deadcode_removal():
#     def f():
#         a = {'what': 'ever'}
#         del a['what']
#         a[1] = 2
#         return len(a)
#         blabla = 5
#         return blabla
#     assert 'blabla' not in decompile(f)

# def test_while2():
#     def f(a):
#         while a < 5:
#             a += 1
#             if a > 3:
#                 continue
#             else:
#                 break
#         return 2 * a
#     scope = {}
#     exec(decompile(f), scope)
#     for a in range(10):
#         assert f(a) == scope['f'](a)

# def test_while_else():
#     def f(a, b):
#         while a < 5 and b < 0:
#             a += 1
#         else:
#             a = 1
#             b += 1
#         return a * b
#     scope = {}
#     exec(decompile(f), scope)
#     for a in [4, 6]:
#         for b in [-3, 4]:
#             assert f(a, b) == scope['f'](a, b)

# def test_rename():
#     def f(a, b):
#         return a * b
#     scope = {}
#     exec(Decompiler(f).decompile(overwite_fn_name="g"), scope)
#     lambda_f = lambda a, b: a * b
#     exec(Decompiler(lambda_f).decompile(overwite_fn_name="lam"), scope)
#     a = 1
#     b = 2
#     assert f(a, b) == scope['g'](a, b)
#     assert f(a, b) == scope['lam'](a, b)


def test_shortcircuit():
    def f(a, b):
        if a > 0 and b > 0:
            return a + b
        elif a > 1 or b > 2:
            return a - b
        else:
            return 2
    scope = {}
    exec(decompile(f), scope)
    for a in [-1, 0, 1, 2]:
        for b in [-1, 0, 1, 2]:
            assert f(a, b) == scope['f'](a, b)



from depyf import decompile
import unittest
import dis
import sys
from dataclasses import dataclass
from copy import deepcopy

@dataclass
class Point:
    x: int
    y: int
    def __matmul__(self, other):
        return self.x * other.x + self.y * other.y

    def __imatmul__(self, other):
        self.x = self.x * other.x
        self.y = self.y * other.y
        return self

    def __setitem__(self, key, value):
        if key == 0:
            self.x = value
        elif key == 1:
            self.y = value
        else:
            raise IndexError("Point only has two dimensions")


point = Point(1, 2)
data_map = {1: 2}


def test_UNARY_POSITIVE():
    def f():
        x = 1
        return +x
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_UNARY_NEGATIVE():
    def f():
        x = 1
        return -x
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_UNARY_NOT():
    def f():
        x = 1
        return not x
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_UNARY_INVERT():
    def f():
        x = 1
        return ~x
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_POWER():
    def f():
        a = 2
        b = 3
        return (a ** b) ** a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_MULTIPLY():
    def f():
        a = 2
        b = 3
        return (a ** b) ** a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_MATRIX_MULTIPLY():
    def f():
        return point @ point
    ans = f()
    scope = {'point': point}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_FLOOR_DIVIDE():
    def f():
        a = 2
        b = 3
        return (a // b) // a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_TRUE_DIVIDE():
    def f():
        a = 2
        b = 3
        return (a / b) / a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_MODULO():
    def f():
        a = 2
        b = 3
        return (a % b) % a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_ADD():
    def f():
        a = 2
        b = 3
        return (a + b) + a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_SUBTRACT():
    def f():
        a = 2
        b = 3
        return (a - b) - a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_SUBSCR():
    def f():
        a = (1, 2, 3)
        return a[0]
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_LSHIFT():
    def f():
        a = 2
        b = 3
        return (a << b) << a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_RSHIFT():
    def f():
        a = 2
        b = 3
        return (a >> b) >> a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_AND():
    def f():
        a = 2
        b = 3
        return (a & b) & a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_XOR():
    def f():
        a = 2
        b = 3
        return (a ^ b) ^ a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BINARY_OR():
    def f():
        a = 2
        b = 3
        return (a | b) | a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_POWER():
    def f():
        a = 2
        a **= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_MULTIPLY():
    def f():
        a = 2
        a *= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_MATRIX_MULTIPLY():
    def f():
        point = Point(1, 2)
        point @= point
        return point
    ans = f()
    scope = {'Point': Point}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_FLOOR_DIVIDE():
    def f():
        a = 2
        a //= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_TRUE_DIVIDE():
    def f():
        a = 2
        a /= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_MODULO():
    def f():
        a = 2
        a %= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_ADD():
    def f():
        a = 2
        a += 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_SUBTRACT():
    def f():
        a = 2
        a -= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_LSHIFT():
    def f():
        a = 2
        a <<= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_RSHIFT():
    def f():
        a = 2
        a >>= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_AND():
    def f():
        a = 2
        a &= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_XOR():
    def f():
        a = 2
        a ^= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_INPLACE_OR():
    def f():
        a = 2
        a |= 3
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_STORE_SUBSCR():
    def f():
        point = Point(1, 2)
        point[0] = 3
        return point
    ans = f()
    scope = {'Point': Point}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_DELETE_SUBSCR():
    def f():
        a = deepcopy(data_map)
        del a[1]
        return a
    ans = f()
    scope = {'data_map': data_map, 'deepcopy': deepcopy}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_GET_LEN():
    def f():
        return len((1, 2, 3))
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_STORE_GLOBAL():
    def f():
        global len
        len = 1
        return len
    ans = f()
    scope = {'len': len}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_DELETE_NAME():
    def f():
        a = 1
        del a
        a = 2
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_UNPACK_SEQUENCE():
    def f():
        a, b = (1, 2)
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_UNPACK_EX():
    def f():
        a, *b = (1, 2, 3)
        return b
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_STORE_ATTR():
    def f():
        point = Point(1, 2)
        point.x = 5
        return point
    ans = f()
    scope = {'Point': Point}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_DELETE_ATTR():
    def f():
        point = Point(1, 2)
        del point.x
        point.x = 9
        return point
    ans = f()
    scope = {'Point': Point}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BUILD_TUPLE():
    def f():
        a = 1
        b = 2
        return (a, b), (a,)
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BUILD_TUPLE_UNPACK():
    def f():
        a = [1, 2, 3]
        b = {4, 5, 6}
        return (*a, *b), (*a,)
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BUILD_LIST_UNPACK():
    def f():
        a = [1, 2, 3]
        b = {4, 5, 6}
        return [*a, *b], [*a,]
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BUILD_SET_UNPACK():
    def f():
        a = [1, 2, 3]
        b = {4, 5, 6}
        return {*a, *b}, {*a,}
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BUILD_MAP_UNPACK():
    def f():
        a = {1: 2}
        b = {3: 4}
        return {**a, **b}, {**a,}
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans



def test_BUILD_LIST():
    def f():
        a = 1
        b = 2
        return [a, b], [a]
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BUILD_SET():
    def f():
        a = 1
        b = 2
        return {a, b}, {a}
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BUILD_MAP():
    def f():
        a = 1
        b = 2
        return {a: 1, 2: 3}, {b: a}
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BUILD_CONST_KEY_MAP():
    def f():
        return {5: 1, 2: 3}
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_LIST_TO_TUPLE():
    # not clear how to test this
    def f():
        return
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_LIST_EXTEND():
    def f():
        return [1, 2, 3]
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_SET_UPDATE():
    def f():
        return {1, 2, 3}
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_DICT_UPDATE():
    def f():
        a = {1: 2}
        b = {'a': 4}
        return {**a, **b}
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_DICT_MERGE():
    def f():
        a = {1: 2}
        b = {'a': 4}
        a.update(**b)
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_CALL_FUNCTION_NORMAL():
    def func(a, b, c=1):
        return (a, b, c)
    def f():
        a = [1, 2, 3]
        b = {'a': 4}
        ans1 = func(1, a, b)
        ans2 = func(b, 1, a)
        ans3 = func(a=a, b=b)
        ans4 = func(a=b, b=a)
        return ans1, ans2, ans3, ans4
    ans = f()
    scope = {'func': func}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_function_signature():
    def func(a, b, c=1, *, d=2):
        return (a, b, c, d)

    scope = {}
    a = [1, 2, 3]
    b = {'a': 4}
    exec(decompile(func), scope)
    assert scope['func'](1, a, b, d=5) == func(1, a, b, d=5)
    assert scope['func'](b, 1, a, d=5) == func(b, 1, a, d=5)
    # assert scope['func'](a=a, b=b, d=5) == func(a=a, b=b, d=5)
    # assert scope['func'](a=b, b=a, d=5) == func(a=b, b=a, d=5)


def test_CALL_FUNCTION_EX():
    def func(*args, **kwargs):
        return (args, kwargs)
    def f():
        a = [1, 2, 3]
        b = {'a': 4}
        ans1 = func(*a)
        ans2 = func(**b)
        ans3 = func(*a, **b)
        ans4 = func()
        ans5 = func(1, 2, 3, *a)
        ans6 = func(1, 2, 3, **b)
        ans7 = func(b=2, **b)
        return ans1, ans2, ans3, ans4, ans5, ans6, ans7
    ans = f()
    scope = {'func': func}
    exec(decompile(f), scope)
    assert scope['f']() == ans

def test_var_args():
    def func(*args, **kwargs):
        return (args, kwargs)
    scope = {}
    exec(decompile(func), scope)
    a = [1, 2, 3]
    b = {'a': 4}
    assert scope['func'](1, a, b, d=5) == func(1, a, b, d=5)


def test_complex_signature():
    def func(a, b, *args, **kwargs):
        return (a, b, args, kwargs)
    scope = {}
    exec(decompile(func), scope)
    a = [1, 2, 3]
    b = {'a': 4}
    assert scope['func'](1, a, b, d=5) == func(1, a, b, d=5)


def test_LOAD_ATTR():
    def f():
        point = Point(1, 2)
        return point.x
    ans = f()
    scope = {'Point': Point}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_COMPARE_OP():
    def f():
        return (3 == 3) + (1 < 2) + (2 > 1) + (2 >= 2) + (1 <= 2) + (1 != 2)
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_IS_OP():
    def f():
        return (int is int), (int is not float)
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_CONTAINS_OP():
    def f():
        return (1 in [1, 2, 3]), (5 not in (6, 7, 4))
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_IMPORT_NAME():
    def f():
        from math import sqrt
        import functools
        return functools.partial(sqrt, 0.3)()
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_BUILD_SLICE():
    def f():
        a = [1, 2, 3]
        return a[:] + a[1:] + a[:2] + a[1:2] + a[::-1]
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans

def test_LIST_COMP():
    def f(a):
        return [i ** 2 for i in range(a)]
    scope = {}
    exec(decompile(f), scope)
    for a in range(10):
        assert scope['f'](a) == f(a)

def test_SET_COMP():
    def f(a):
        return {i ** 2 for i in range(a)}
    scope = {}
    exec(decompile(f), scope)
    for a in range(10):
        assert scope['f'](a) == f(a)

def test_MAP_COMP():
    def f(a):
        return {i: i ** 2 for i in range(a)}
    scope = {}
    exec(decompile(f), scope)
    for a in range(10):
        assert scope['f'](a) == f(a)

def test_NESTED_COMP():
    def f(a):
        return [{x: {_ for _ in range(x)} for x in range(i)} for i in range(a)]
    scope = {}
    exec(decompile(f), scope)
    for a in range(10):
        assert scope['f'](a) == f(a)


def test_FORMAT_VALUE():
    def f():
        a = 1
        b = 2
        c = 3
        return f"{a} {b!r} {b!s} {b!a} {c:.2f}"
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_ROT_TWO():
    def f():
        a = 1
        b = 2
        c = 3
        d = 4
        e = 5
        a, b = b, a
        a, b, c = c, b, a
        a, b, c, d = d, c, b, a
        a, b, c, d, e = e, d, c, b, a
        return a
    ans = f()
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == ans


def test_IF():
    def f(a):
        if a == 0:
            return 0
        elif a == 1:
            return 1
        else:
            return 2
    scope = {}
    exec(decompile(f), scope)
    for a in range(10):
        assert scope['f'](a) == f(a)

def test_compound_IF_and():
    def f(a, b):
        c = 1
        if a > 0 and b > 1:
            c += 1
        else:
            c += 2
        c += 3
        return c
    scope = {}
    exec(decompile(f), scope)
    for a in range(-3, 3):
        for b in range(-3, 3):
            assert scope['f'](a, b) == f(a, b)

def test_compound_IF_or():
    def f(a, b):
        c = 1
        if a > 0 or b > 1:
            c += 1
        else:
            c += 2
        c += 3
        return c
    scope = {}
    exec(decompile(f), scope)
    for a in range(-3, 3):
        for b in range(-3, 3):
            assert scope['f'](a, b) == f(a, b)

def test_IF_NONE():
    def f(a):
        if a is None:
            return 0
        elif a is not None:
            return 1
    scope = {}
    exec(decompile(f), scope)
    for a in range(10):
        assert scope['f'](a) == f(a)

def test_MAKE_FUNCTION():
    def f(a):
        def g(b=3):
            return a + b
        return g(2)
    scope = {}
    exec(decompile(f), scope)
    for a in range(10):
        assert scope['f'](a) == f(a)

def test_simple_try():
    def f(a):
        try:
            a += 1
        finally:
            a += 2
        a += 3
        return a
    scope = {}
    exec(decompile(f), scope)
    for a in range(10):
        assert scope['f'](a) == f(a)

def test_simple_for():
    def f(a):
        for i in range(5):
            a += i
        return a
    scope = {}
    exec(decompile(f), scope)
    for a in range(10):
        assert scope['f'](a) == f(a)

class A:
    def f(self):
        return __class__

def test_class_method():
    scope = {}
    for k, v in zip(A.f.__code__.co_freevars, A.f.__closure__):
        scope[k] = v.cell_contents
    exec(decompile(A.f), scope)
    a = A()
    assert scope['f'](a) == a.f()


def test_EXTENDED_ARG():
    def f():
        a_658 = 658; a_358 = 358; a_308 = 308; a_344 = 344; a_640 = 640; a_256 = 256; a_347 = 347; a_424 = 424; a_336 = 336; a_539 = 539; a_56 = 56; a_680 = 680; a_860 = 860; a_580 = 580; a_230 = 230; a_992 = 992; a_433 = 433; a_892 = 892; a_888 = 888; a_841 = 841; a_748 = 748; a_690 = 690; a_166 = 166; a_118 = 118; a_165 = 165; a_842 = 842; a_575 = 575; a_556 = 556; a_402 = 402; a_640 = 640; a_154 = 154; a_979 = 979; a_350 = 350; a_378 = 378; a_742 = 742; a_579 = 579; a_830 = 830; a_748 = 748; a_688 = 688; a_884 = 884; a_118 = 118; a_815 = 815; a_413 = 413; a_963 = 963; a_864 = 864; a_296 = 296; a_552 = 552; a_245 = 245; a_969 = 969; a_999 = 999; a_71 = 71; a_83 = 83; a_964 = 964; a_169 = 169; a_48 = 48; a_151 = 151; a_100 = 100; a_30 = 30; a_478 = 478; a_458 = 458; a_233 = 233; a_800 = 800; a_191 = 191; a_286 = 286; a_488 = 488; a_101 = 101; a_456 = 456; a_478 = 478; a_347 = 347; a_853 = 853; a_961 = 961; a_572 = 572; a_407 = 407; a_353 = 353; a_79 = 79; a_514 = 514; a_801 = 801; a_382 = 382; a_898 = 898; a_979 = 979; a_696 = 696; a_751 = 751; a_367 = 367; a_380 = 380; a_52 = 52; a_445 = 445; a_321 = 321; a_728 = 728; a_19 = 19; a_64 = 64; a_679 = 679; a_46 = 46; a_402 = 402; a_199 = 199; a_479 = 479; a_370 = 370; a_768 = 768; a_988 = 988; a_205 = 205; a_19 = 19; a_125 = 125; a_821 = 821; a_335 = 335; a_816 = 816; a_135 = 135; a_210 = 210; a_212 = 212; a_926 = 926; a_726 = 726; a_384 = 384; a_279 = 279; a_157 = 157; a_457 = 457; a_595 = 595; a_184 = 184; a_410 = 410; a_375 = 375; a_981 = 981; a_154 = 154; a_489 = 489; a_209 = 209; a_186 = 186; a_784 = 784; a_317 = 317; a_110 = 110; a_251 = 251; a_752 = 752; a_138 = 138; a_119 = 119; a_100 = 100; a_74 = 74; a_951 = 951; a_3 = 3; a_987 = 987; a_655 = 655; a_422 = 422; a_89 = 89; a_930 = 930; a_235 = 235; a_973 = 973; a_391 = 391; a_956 = 956; a_641 = 641; a_658 = 658; a_458 = 458; a_883 = 883; a_440 = 440; a_369 = 369; a_99 = 99; a_503 = 503; a_94 = 94; a_195 = 195; a_403 = 403; a_182 = 182; a_192 = 192; a_196 = 196; a_903 = 903; a_108 = 108; a_224 = 224; a_891 = 891; a_922 = 922; a_278 = 278; a_201 = 201; a_485 = 485; a_200 = 200; a_28 = 28; a_249 = 249; a_578 = 578; a_857 = 857; a_411 = 411; a_807 = 807; a_267 = 267; a_667 = 667; a_53 = 53; a_907 = 907; a_26 = 26; a_273 = 273; a_672 = 672; a_945 = 945; a_931 = 931; a_318 = 318; a_795 = 795; a_439 = 439; a_965 = 965; a_939 = 939; a_891 = 891; a_572 = 572; a_283 = 283; a_706 = 706; a_124 = 124; a_383 = 383; a_615 = 615; a_883 = 883; a_348 = 348; a_805 = 805; a_9 = 9; a_970 = 970; a_698 = 698; a_526 = 526; a_125 = 125; a_621 = 621; a_826 = 826; a_273 = 273; a_340 = 340; a_781 = 781; a_581 = 581; a_112 = 112; a_417 = 417; a_632 = 632; a_277 = 277; a_432 = 432; a_984 = 984; a_437 = 437; a_795 = 795; a_437 = 437; a_651 = 651; a_288 = 288; a_369 = 369; a_722 = 722; a_630 = 630; a_65 = 65; a_470 = 470; a_307 = 307; a_913 = 913; a_666 = 666; a_88 = 88; a_326 = 326; a_342 = 342; a_912 = 912; a_756 = 756; a_540 = 540; a_84 = 84; a_595 = 595; a_769 = 769; a_607 = 607; a_130 = 130; a_556 = 556; a_359 = 359; a_237 = 237; a_982 = 982; a_274 = 274; a_796 = 796; a_604 = 604; a_232 = 232; a_559 = 559; a_629 = 629; a_329 = 329; a_919 = 919; a_61 = 61; a_642 = 642; a_127 = 127; a_355 = 355; a_694 = 694; a_199 = 199; a_261 = 261; a_348 = 348; a_319 = 319; a_753 = 753; a_621 = 621; a_916 = 916; a_601 = 601; a_125 = 125; a_854 = 854; a_539 = 539; a_212 = 212; a_297 = 297; a_736 = 736; a_365 = 365; a_965 = 965; a_534 = 534; a_472 = 472; a_302 = 302; a_513 = 513; a_450 = 450; a_370 = 370; a_346 = 346; a_492 = 492; a_112 = 112; a_457 = 457; a_155 = 155; a_423 = 423; a_782 = 782; a_941 = 941; a_860 = 860; a_932 = 932; a_359 = 359; a_329 = 329; a_554 = 554; a_299 = 299; a_456 = 456; a_663 = 663; a_933 = 933; a_826 = 826; a_715 = 715; a_319 = 319; a_178 = 178; a_551 = 551; a_671 = 671; a_41 = 41; a_702 = 702; 
        a = 10
        return a
    scope = {}
    exec(decompile(f), scope)
    assert scope['f']() == f()



"""A simple program to transform bytecode into more readable source code."""

import sys
import os
import dis
from types import CodeType
from typing import List, Tuple, Dict, Union, Callable, Optional
import dataclasses
import inspect
import functools
from collections import defaultdict
import contextlib

from .code_transform import (
    nop_unreachable_bytecode,
    add_indentation,
    remove_indentation,
    remove_some_temp,
    propagate_line_nums,
    convert_instruction,
    simplify_with_statement,
    simplify_finally_statement,
    Instruction,
)
from .utils import (
    get_function_signature,
)


@dataclasses.dataclass
class DecompilerState:
    """State of decompiler, keep track of the evaluation stack, as well as the decompiled source code."""
    source_code: str
    stack: list


@dataclasses.dataclass
class Decompiler:
    """A decompiler for a code object."""
    code: CodeType
    temp_count: int = 0
    temp_prefix: str = "__temp_"
    state: DecompilerState = dataclasses.field(default_factory=lambda: DecompilerState(source_code="", stack=[]))
    indentation: int = 4

    @contextlib.contextmanager
    def new_state(self, stack):
        """Create a new state for decompiler."""
        state = DecompilerState(source_code="", stack=stack)
        old_state = self.state
        self.state = state
        yield
        self.state = old_state

# ==================== Load Instructions =============================

    def LOAD_CONST(self, inst: Instruction):
        """Push a constant onto the stack.
        `inst.argval` is the constant value, we have to use `repr` to get the source code
        """
        can_repr = False
        try:
            can_repr = eval(repr(inst.argval)) == inst.argval
        except:
            pass
        if can_repr:
            self.state.stack.append(repr(inst.argval))
        else:
            if isinstance(inst.argval, type):
                # Don't know why a class type get here, support this corner case anyway.
                module = inst.argval.__module__
                name = inst.argval.__name__
                self.state.source_code += "import importlib\n"
                temp_name = self.get_temp_name()
                self.state.source_code += f'{temp_name} = importlib.import_module("{module}").{name}\n'
                self.state.stack.append(temp_name)
            elif inst.argrepr.startswith("torch."):
                # Don't know why torch.xxx get here, support this corner case anyway. This deals with something like `torch.float`.
                self.state.source_code += "import torch\n"
                temp_name = self.get_temp_name()
                self.state.source_code += f'{temp_name} = {inst.argval}\n'
                self.state.stack.append(temp_name)
            else:
                self.state.stack.append(inst.argval)

    def generic_load(self, inst: Instruction):
        """`inst.argval` is the variable name, in string"""
        if "NULL + " in inst.argrepr:
            # Python 3.11 support
            self.state.stack.append(None)
        if inst.argrepr.startswith("."):
            # list/set/tuple comprehension.
            self.state.stack.append(inst.argval.replace(".", "comp_arg_"))
        else:
            self.state.stack.append(inst.argval)

    LOAD_FAST = LOAD_FAST_AND_CLEAR = LOAD_FAST_CHECK = LOAD_GLOBAL = LOAD_DEREF = LOAD_NAME = LOAD_CLASSDEREF = LOAD_CLOSURE = generic_load

    def LOAD_LOCALS(self, inst: Instruction):
        self.state.stack.append("locals()")
        self.replace_mutable_tos_with_temp()

    def LOAD_FROM_DICT_OR_GLOBALS(self, inst: Instruction):
        tos = self.state.stack.pop()
        self.state.stack.append(f"{tos}[{inst.argval}] if '{inst.argval}' in {tos} else {inst.argval}")
        self.replace_mutable_tos_with_temp()

    LOAD_FROM_DICT_OR_DEREF = LOAD_FROM_DICT_OR_GLOBALS

    def MAKE_FUNCTION(self, inst: Instruction):
        if sys.version_info < (3, 11):
            qual_name = self.state.stack.pop()
            try:
                qual_name = eval(qual_name)
            except Exception:
                pass
            func_name = qual_name
            if "<" in func_name:
                self.state.source_code += f'"original name {qual_name} is illegal, use a temp name."\n'
                func_name = self.get_temp_name()
        else:
            # Python 3.11 support, see https://docs.python.org/3.11/library/dis.html#opcode-MAKE_FUNCTION
            func_name = self.get_temp_name()
        code = self.state.stack.pop()
        if inst.argval & 0x08:
            # has closure
            self.state.stack.pop()
        if inst.argval & 0x04:
            # has annotations
            self.state.stack.pop()
        kw_defaults = self.state.stack.pop() if inst.argval & 0x02 else {}
        defaults = self.state.stack.pop() if inst.argval & 0x01 else ()
        if len(kw_defaults) or len(defaults):
            print("Function with default arguments is not supported, ignore the default arguments")
        this_index = self.index_of(inst.offset)
        immediately_used = False
        if self.instructions[this_index + 1].opname == "STORE_FAST":
            # the function is immediately stored in a variable, use that variable name
            func_name = self.instructions[this_index + 1].argval
            immediately_used = True
        inner_func = Decompiler(code).decompile(overwite_fn_name=func_name)
        self.state.source_code += inner_func
        if not immediately_used:
            self.state.stack.append(func_name)
        else:
            # skip one instruction
            return this_index + 2

    def COPY_FREE_VARS(self, inst: Instruction):
        # this opcode is used to copy free variables from the outer scope to the closure
        # it affects the frame, but not the stack or the source code
        pass

    def LOAD_ATTR(self, inst: Instruction):
        self.state.stack.append(f"getattr({self.state.stack.pop()}, {repr(inst.argval)})")

    def LOAD_SUPER_ATTR(self, inst: Instruction):
        # not tested
        self_obj = self.state.stack.pop()
        cls_obj = self.state.stack.pop()
        super_obj = self.state.stack.pop()
        self.state.stack.append(f"{super_obj}({cls_obj}, {self_obj}).{inst.argval}")
        self.replace_mutable_tos_with_temp()

    def LOAD_METHOD(self, inst: Instruction):
        self.state.stack.append(f"{self.state.stack.pop()}.{inst.argval}")

    def LOAD_ASSERTION_ERROR(self, inst: Instruction):
        self.state.stack.append("AssertionError")

    def PUSH_NULL(self, inst: Instruction):
        # the `None` object is used to represent `NULL` in python bytecode
        self.state.stack.append(None)

    def GET_ITER(self, inst: Instruction):
        tos = self.state.stack.pop()
        self.state.stack.append(f"iter({tos})")

# ==================== Store Instructions =============================

    def generic_store(self, inst: Instruction):
        left = inst.argval
        right = self.state.stack.pop()
        if left != right:
            # Inplace operations like `+=` will pop the variable name from the stack, and push the result back to the stack
            # leading to a source code like `x = x`. We need to avoid this.
            self.state.source_code += f"{left} = {right}\n"

    STORE_FAST = STORE_GLOBAL = STORE_DEREF = STORE_NAME = generic_store

    def STORE_SUBSCR(self, inst: Instruction):
        index = self.state.stack.pop()
        x = self.state.stack.pop()
        value = self.state.stack.pop()
        self.state.source_code += f"{x}[{index}] = {value}\n"

    def STORE_SLICE(self, inst: Instruction):
        # not tested, code according to https://docs.python.org/3.12/library/dis.html#opcode-STORE_SLICE
        end = self.state.stack.pop()
        start = self.state.stack.pop()
        container = self.state.stack.pop()
        value = self.state.stack.pop()
        self.state.source_code += f"{container}[{start}:{end}] = {value}\n"

    def STORE_ATTR(self, inst: Instruction):
        x = self.state.stack.pop()
        value = self.state.stack.pop()
        self.state.source_code += f"{x}.{inst.argval} = {value}\n"

# ==================== Del Instructions =============================

    def DELETE_SUBSCR(self, inst: Instruction):
        index = self.state.stack.pop()
        x = self.state.stack.pop()
        self.state.source_code += f"del {x}[{index}]\n"

    def generic_delete(self, inst: Instruction):
        self.state.source_code += f"del {inst.argval}\n"
    
    DELETE_NAME = DELETE_FAST = DELETE_GLOBAL = DELETE_DEREF = generic_delete

    def DELETE_ATTR(self, inst: Instruction):
        x = self.state.stack.pop()
        self.state.source_code += f"del {x}.{inst.argval}\n"

# ==================== Import Instructions =============================
    def IMPORT_NAME(self, inst: Instruction):
        # TODO: check multi-level import, e.g. `import a.b.c`
        name = inst.argval.split(".")[0]
        fromlist = self.state.stack.pop()
        level = self.state.stack.pop()
        self.state.source_code += f"{name} = __import__({repr(inst.argval)}, fromlist={fromlist}, level={level})\n"
        self.state.stack.append(name)
    
    def IMPORT_FROM(self, inst: Instruction):
        name = inst.argval
        module = self.state.stack[-1]
        self.state.source_code += f"{name} = {module}.{name}\n"
        self.state.stack.append(name)

# ==================== Unary Instructions =============================

    def generic_unary(self, inst: Instruction):
        op = {
            "UNARY_NEGATIVE": "-",
            "UNARY_POSITIVE": "+",
            "UNARY_INVERT": "~",
            "UNARY_NOT": "not",
        }[inst.opname]
        self.state.stack.append(f"({op} {self.state.stack.pop()})")
    
    UNARY_NEGATIVE = UNARY_POSITIVE = UNARY_INVERT = UNARY_NOT = generic_unary

    def GET_LEN(self, inst: Instruction):
        self.state.stack.append(f"len({self.state.stack[-1]})")

# ==================== Binary Instructions =============================
    def generic_binary(self, inst: Instruction):
        rhs = self.state.stack.pop()
        lhs = self.state.stack.pop()
        op = {
            "BINARY_MULTIPLY": "*",
            "BINARY_ADD": "+",
            "BINARY_SUBTRACT": "-",
            "BINARY_TRUE_DIVIDE": "/",
            "BINARY_FLOOR_DIVIDE": "//",
            "BINARY_MODULO": "%",
            "BINARY_POWER": "**",
            "BINARY_AND": "&",
            "BINARY_OR": "|",
            "BINARY_XOR": "^",
            "BINARY_LSHIFT": "<<",
            "BINARY_RSHIFT": ">>",
            "BINARY_MATRIX_MULTIPLY": "@",
        }[inst.opname]
        self.state.stack.append(f"({lhs} {op} {rhs})")

    BINARY_MULTIPLY = BINARY_ADD = BINARY_SUBTRACT = BINARY_TRUE_DIVIDE = BINARY_FLOOR_DIVIDE = BINARY_MODULO = BINARY_POWER = BINARY_AND = BINARY_OR = BINARY_XOR = BINARY_LSHIFT = BINARY_RSHIFT = BINARY_MATRIX_MULTIPLY = generic_binary

    def BINARY_SUBSCR(self, inst: Instruction):
        rhs = self.state.stack.pop()
        lhs = self.state.stack.pop()
        self.state.stack.append(f"{lhs}[{rhs}]")

    def BINARY_SLICE(self, inst: Instruction):
        end = self.state.stack.pop()
        start = self.state.stack.pop()
        container = self.state.stack.pop()
        self.state.stack.append(f"{container}[{start}:{end}]")

# ==================== Binary Inplace Instructions =============================
    def generic_inplace_binary(self, inst: Instruction):
        rhs = self.state.stack.pop()
        lhs = self.state.stack.pop()
        op = {
            "INPLACE_MULTIPLY": "*",
            "INPLACE_ADD": "+",
            "INPLACE_SUBTRACT": "-",
            "INPLACE_TRUE_DIVIDE": "/",
            "INPLACE_FLOOR_DIVIDE": "//",
            "INPLACE_MODULO": "%",
            "INPLACE_POWER": "**",
            "INPLACE_AND": "&",
            "INPLACE_OR": "|",
            "INPLACE_XOR": "^",
            "INPLACE_LSHIFT": "<<",
            "INPLACE_RSHIFT": ">>",
            "INPLACE_MATRIX_MULTIPLY": "@",
        }[inst.opname]
        self.state.source_code += f"{lhs} {op}= {rhs}\n"
        self.state.stack.append(lhs)
    
    INPLACE_MULTIPLY = INPLACE_ADD = INPLACE_SUBTRACT = INPLACE_TRUE_DIVIDE = INPLACE_FLOOR_DIVIDE = INPLACE_MODULO = INPLACE_POWER = INPLACE_AND = INPLACE_OR = INPLACE_XOR = INPLACE_LSHIFT = INPLACE_RSHIFT = INPLACE_MATRIX_MULTIPLY = generic_inplace_binary

    def BINARY_OP(self, inst: Instruction):
        rhs = self.state.stack.pop()
        lhs = self.state.stack.pop()
        if "=" in inst.argrepr:
            self.state.source_code += f"{lhs} {inst.argrepr} {rhs}\n"
            self.state.stack.append(lhs)
        else:
            self.state.stack.append(f"({lhs} {inst.argrepr} {rhs})")

# ==================== Conditional Test Instructions =============================
    def COMPARE_OP(self, inst: Instruction):
        rhs = self.state.stack.pop()
        lhs = self.state.stack.pop()
        self.state.stack.append(f"({lhs} {inst.argval} {rhs})")

    def IS_OP(self, inst: Instruction):
        rhs = self.state.stack.pop()
        lhs = self.state.stack.pop()
        op = "is" if inst.argval == 0 else "is not"
        self.state.stack.append(f"({lhs} {op} {rhs})")

    def CONTAINS_OP(self, inst: Instruction):
        rhs = self.state.stack.pop()
        lhs = self.state.stack.pop()
        op = "in" if inst.argval == 0 else "not in"
        self.state.stack.append(f"({lhs} {op} {rhs})")

# ==================== Control Flow Instructions =============================

    def BREAK_LOOP(self, inst: Instruction):
        self.state.source_code += "break\n"
    
    def generic_abs_jump(self, inst: Instruction):
        jump_offset = inst.get_jump_target()
        if jump_offset > inst.offset:
            self.state.source_code += "break\n"
        else:
            self.state.source_code += "continue\n"
    
    JUMP_ABSOLUTE = JUMP_FORWARD = JUMP_BACKWARD = JUMP_BACKWARD_NO_INTERRUPT = generic_abs_jump

    def RETURN_VALUE(self, inst: Instruction):
        self.state.source_code += f"return {self.state.stack[-1]}\n"

    def RETURN_CONST(self, inst: Instruction):
        self.state.source_code += f"return {inst.argval}\n"

    def YIELD_VALUE(self, inst: Instruction):
        if sys.version_info >= (3, 12):
            raise NotImplementedError("YIELD_VALUE is not supported in Python 3.12")
        self.state.source_code += f"yield {self.state.stack[-1]}\n"

    def RETURN_GENERATOR(self, inst: Instruction):
        # we don't handle generator/coroutine, add this to support simple yield
        self.state.stack.append(None)
    def GEN_START(self, inst: Instruction):
        # self.state.stack.pop()
        assert inst.argval == 0, "Only generator expression is supported"

    def generic_jump_if(self, inst: Instruction):
        """Only support consecutive and/or, do not support mixed and/or."""
        jump_offset = inst.get_jump_target()
        jump_index = self.index_of(jump_offset)
        this_index = self.index_of(inst.offset)
        cond = self.state.stack[-1]

        if "ASSERT" in self.instructions[this_index + 1].opname:
            with self.new_state(self.state.stack):
                self.decompile_range(this_index + 1, jump_index)
                source_code = self.state.source_code
            source_code = add_indentation(source_code, self.indentation)
            self.state.source_code += f"if not {cond}:\n{source_code}"
            return jump_index

        fallthrough_stack = self.state.stack

        if_body_start_offset = None
        if_body_end_offset = None
        last_index = this_index
        conditions = [cond]
        for _index in range(this_index, jump_index):
            _inst = self.instructions[_index]
            if "IF_FALSE" in _inst.opname or "IF_NOT_NONE" in _inst.opname or "IF_NONE" in _inst.opname:
                # JUMP_IF_FALSE, followed by "and", short-circuit evaluation means we jump to the end of if-block if the condition is false
                if if_body_end_offset is None:
                    if_body_end_offset = _inst.get_jump_target()
                if _inst.get_jump_target() == if_body_end_offset:
                    if _index != this_index:
                        with self.new_state(fallthrough_stack):
                            self.decompile_range(last_index + 1, _index)
                            source_code = self.state.source_code
                        self.state.source_code += source_code
                        conditions.append(self.state.stack[-1])
                        last_index = _index
                    if "IF_NOT_NONE" in _inst.opname:
                        conditions[-1] = f"{conditions[-1]} is None"
                    elif "IF_NONE" in _inst.opname:
                        conditions[-1] = f"{conditions[-1]} is not None"
                    conditions.append("and")

                    jump_stack = fallthrough_stack.copy()
                    fallthrough_stack.pop()
                    # POP_AND_JUMP / JUMP_OR_POP
                    if "POP_JUMP" in _inst.opname:
                        jump_stack.pop()
                    elif "OR_POP" in _inst.opname:
                        pass

            elif "IF_TRUE" in _inst.opname:
                # JUMP_IF_TRUE, followed by "or", short-circuit evaluation means we jump to the start of if-block if the condition is true
                if if_body_start_offset is None:
                    if_body_start_offset = _inst.get_jump_target()
                if _inst.get_jump_target() == if_body_start_offset:
                    if _index != this_index:
                        with self.new_state(fallthrough_stack):
                            self.decompile_range(last_index + 1, _index)
                            source_code = self.state.source_code
                        self.state.source_code += source_code
                        conditions.append(self.state.stack[-1])
                        last_index = _index
                    conditions.append("or")

                    jump_stack = fallthrough_stack.copy()
                    fallthrough_stack.pop()
                    # POP_AND_JUMP / JUMP_OR_POP
                    if "POP_JUMP" in _inst.opname:
                        jump_stack.pop()
                    elif "OR_POP" in _inst.opname:
                        pass

        conditions.pop()

        if if_body_start_offset is None:
            if_body_start_offset = self.instructions[last_index + 1].offset
            
        if_body_start = self.index_of(if_body_start_offset)

        if if_body_end_offset is None:
            # Don't know where the if body ends, so we have to find the next jump instruction
            if_body_end = if_body_start + 1
            while if_body_end < len(self.instructions) and not self.instructions[if_body_end].is_jump():
                if_body_end += 1
            if if_body_end == len(self.instructions):
                if_body_end -= 1
            if_body_end_offset = self.instructions[if_body_end].offset
        else:
            if_body_end = self.index_of(if_body_end_offset)
        if jump_index < if_body_start:
            self.state.source_code += add_indentation("continue\n", self.indentation)
            return

        jump_targets = [i.get_jump_target() for i in self.instructions[if_body_start + 1: if_body_end] if i.is_jump() and i.get_jump_target() > if_body_end_offset]
        else_code = ""
        if jump_targets:
            # has "else" branch
            else_start_offset = if_body_end_offset
            else_start_index = self.index_of(else_start_offset)
            max_jump = max(jump_targets)
            max_jump_index = self.index_of(max_jump)
            else_code += "else:\n"
            with self.new_state(jump_stack):
                self.decompile_range(else_start_index, max_jump_index)
                source_code = self.state.source_code
            else_code += add_indentation(source_code, self.indentation)

        with self.new_state(fallthrough_stack):
            if else_code and self.instructions[if_body_end - 1].is_jump():
                # the last instruction is a jump, so it is not part of the if body, but the jump out of the if-else block
                if_body_end -= 1
            self.decompile_range(if_body_start, if_body_end)
            if_code = "if " + " ".join(conditions) + ":\n"
            if_code = if_code + add_indentation(self.state.source_code, self.indentation)
        
        self.state.source_code += if_code + else_code
        self.state.stack = fallthrough_stack

        return max_jump_index if else_code else jump_index

    POP_JUMP_IF_TRUE = POP_JUMP_IF_FALSE = generic_jump_if
    POP_JUMP_FORWARD_IF_TRUE = POP_JUMP_FORWARD_IF_FALSE = generic_jump_if
    POP_JUMP_BACKWARD_IF_TRUE = POP_JUMP_BACKWARD_IF_FALSE = generic_jump_if
    POP_JUMP_FORWARD_IF_NONE = POP_JUMP_FORWARD_IF_NOT_NONE = generic_jump_if
    POP_JUMP_BACKWARD_IF_NONE = POP_JUMP_BACKWARD_IF_NOT_NONE = generic_jump_if
    JUMP_IF_TRUE_OR_POP = JUMP_IF_FALSE_OR_POP = generic_jump_if
    POP_JUMP_IF_NOT_NONE = POP_JUMP_BACKWARD_IF_NOT_NONE
    POP_JUMP_IF_NONE = POP_JUMP_BACKWARD_IF_NONE

    def SETUP_FINALLY(self, inst: Instruction):
        start_index = self.index_of(inst.offset)
        end_index = self.index_of(inst.get_jump_target())
        pop_block_index = [i for i, x in enumerate(self.instructions) if x.opname == "POP_BLOCK" and start_index <= i < end_index][-1]

        try_code = ""
        with self.new_state(self.state.stack):
            self.decompile_range(start_index + 1, pop_block_index)
            try_code = self.state.source_code
            try_code = add_indentation(try_code, self.indentation)
            try_code = "try:\n" + try_code
        
        finally_code = ""
        with self.new_state(self.state.stack):
            end_finally_index = [i for i, x in enumerate(self.instructions) if x.opname == "END_FINALLY" and start_index <= i]
            if end_finally_index:
                end_index = end_finally_index[0]
            finally_end_index = end_index
            if self.instructions[finally_end_index - 1].is_jump():
                finally_end_index -= 1
            self.decompile_range(pop_block_index + 1, finally_end_index)
            finally_code = self.state.source_code
            finally_code = add_indentation(finally_code, self.indentation)
            finally_code = "finally:\n" + finally_code
        
        self.state.source_code += try_code + finally_code
        return end_index

    def FOR_ITER(self, inst: Instruction):
        start_index = self.index_of(inst.offset)
        end_index = self.index_of(inst.get_jump_target())

        temp_name = self.get_temp_name()
        for_code = f"for {temp_name} in {self.state.stack.pop()}:\n"
        self.state.stack.append(temp_name)
        with self.new_state(self.state.stack):
            self.decompile_range(start_index + 1, end_index)
            code = self.state.source_code
            for_code = for_code + add_indentation(code, self.indentation)

        self.state.source_code += for_code
        return end_index

# ==================== Stack Manipulation Instructions =============================
    def rot_n(self, inst: Instruction):
        if inst.opname == "ROT_N":
            n = inst.argval
        else:
            n = {
                "ROT_TWO": 2,
                "ROT_THREE": 3,
                "ROT_FOUR": 4,
            }[inst.opname]
        values = self.state.stack[-n:]
        values = [values[-1]] + values[:-1]
        self.state.stack[-n:] = values
        
    ROT_N = ROT_TWO = ROT_THREE = ROT_FOUR = rot_n

    def SWAP(self, inst: Instruction):
        n = inst.argval
        tos = self.state.stack[-1]
        value = self.state.stack[- n]
        tos, value = value, tos
        self.state.stack[-1] = tos
        self.state.stack[- n] = value
    
    def COPY(self, inst: Instruction):
        # not tested, don't know how to generate this instruction
        n = inst.argval
        value = self.state.stack[-1 - n]
        self.state.stack.append(value)
    
    def POP_TOP(self, inst: Instruction):
        self.state.stack.pop()

    def DUP_TOP(self, inst: Instruction):
        # not tested
        self.state.stack.append(self.state.stack[-1])
    
    def DUP_TOP_TWO(self, inst: Instruction):
        # not tested
        tos = self.state.stack[-1]
        tos1 = self.state.stack[-2]
        self.state.stack.append(tos1)
        self.state.stack.append(tos)

# ==================== Function Call Instructions =============================
    def KW_NAMES(self, inst: Instruction):
        names = self.code.co_consts[inst.arg]
        self.state.stack.append(repr(names))

    def CALL(self, inst: Instruction):
        last_inst = [x for x in self.instructions if x.offset < inst.offset]
        has_kw_names = False
        if last_inst:
            if last_inst[-1].opname == "KW_NAMES" or (len(last_inst) > 1 and last_inst[-2].opname == "KW_NAMES" and last_inst[-1].opname == "PRECALL"):
                has_kw_names = True
        kw_names = tuple()
        if has_kw_names:
            kw_names = eval(self.state.stack.pop())
        args = [(self.state.stack.pop()) for _ in range(inst.argval)]
        args = args[::-1]
        pos_args = args[:len(args) - len(kw_names)]
        kwargs = args[len(args) - len(kw_names):]
        kwcalls = []
        for name, value in zip(kwargs, kw_names):
            kwcalls.append(f"{name}={value}")
        func = self.state.stack.pop()
        if self.state.stack and self.state.stack[-1] is None:
            self.state.stack.pop()
        if "iter(" in func:
            # Why do we need this? Don't know. But sometimes CPython generates CALL with argval=0, but the function actually needs an arg (for list/set/map comprehension).
            pos_args = [func]
            func = self.state.stack.pop()
        self.state.stack.append(f"{func}({', '.join(pos_args + kwcalls)})")
        self.replace_mutable_tos_with_temp()

    def generic_call(self, inst: Instruction):
        args = [(self.state.stack.pop()) for _ in range(inst.argval)]
        args = args[::-1]
        func = self.state.stack.pop()
        self.state.stack.append(f"{func}({', '.join(args)})")
        self.replace_mutable_tos_with_temp()
    
    CALL_FUNCTION = CALL_METHOD = generic_call

    def CALL_FUNCTION_KW(self, inst: Instruction):
        kw_args = eval(self.state.stack.pop())
        kw_vals = [(self.state.stack.pop()) for _ in range(len(kw_args))]
        kw_vals.reverse()
        kwcalls = []
        for name, val in zip(kw_args, kw_vals):
            kwcalls.append(f"{name}={val}")
        pos_args = [(self.state.stack.pop()) for _ in range(inst.argval - len(kw_args))]
        pos_args = pos_args[::-1]
        func = self.state.stack.pop()
        self.state.stack.append(f"{func}({', '.join(pos_args + kwcalls)})")
        self.replace_mutable_tos_with_temp()

    def CALL_FUNCTION_EX(self, inst: Instruction):
        if inst.argval == 0:
            args = self.state.stack.pop()
            func = self.state.stack.pop()
            self.state.stack.append(f"{func}(*{args})")
        elif inst.argval == 1:
            kw_args = self.state.stack.pop()
            args = self.state.stack.pop()
            func = self.state.stack.pop()
            self.state.stack.append(f"{func}(*{args}, **{kw_args})")
        self.replace_mutable_tos_with_temp()

    def CALL_INTRINSIC_1(self, inst: Instruction):
        if inst.argrepr in ["INTRINSIC_1_INVALID", "INTRINSIC_IMPORT_STAR", "INTRINSIC_STOPITERATION_ERROR", "INTRINSIC_ASYNC_GEN_WRAP"]:
            # invalid intrinsic, skip
            pass
        elif inst.argrepr in ["INTRINSIC_TYPEVAR", "INTRINSIC_PARAMSPEC", "INTRINSIC_TYPEVARTUPLE", "INTRINSIC_SUBSCRIPT_GENERIC", "INTRINSIC_TYPEALIAS"]:
            # not tested, skip
            pass
        elif inst.argrepr == "INTRINSIC_PRINT":
            self.state.source_code += f"print({self.state.stack.pop()})\n"
            self.state.stack.append("None")
        elif inst.argrepr == "INTRINSIC_UNARY_POSITIVE":
            self.state.stack[-1] = f"+{self.state.stack[-1]}"
        elif inst.argrepr == "INTRINSIC_LIST_TO_TUPLE":
            return self.LIST_TO_TUPLE(inst)


# ==================== Container Related Instructions (tuple, list, set, dict) =============================
    def UNPACK_SEQUENCE(self, inst: Instruction):
        varname = self.state.stack.pop()
        for i in range(inst.argval):
            self.state.stack.append(f"{varname}[{inst.argval - 1 - i}]")
    
    def UNPACK_EX(self, inst: Instruction):
        varname = self.state.stack.pop()
        self.state.stack.append(f"list({varname}[{inst.argval}:])")
        for i in range(inst.argval):
            self.state.stack.append(f"{varname}[{inst.argval - 1 - i}]")

    def BUILD_SLICE(self, inst: Instruction):
        tos = self.state.stack.pop()
        tos1 = self.state.stack.pop()
        if inst.argval == 2:
            self.state.stack.append(f"slice({tos1}, {tos})")
        elif inst.argval == 3:
            tos2 = self.state.stack.pop()
            self.state.stack.append(f"slice({tos2}, {tos1}, {tos})")

    def build_tuple(self, inst: Instruction):
        args = [self.state.stack.pop() for _ in range(inst.argval)]
        args = args[::-1]
        if "UNPACK" in inst.opname:
            args = [f"*{arg}" for arg in args]
        if inst.argval == 1:
            self.state.stack.append(f"({args[0]},)")
        else:
            self.state.stack.append(f"({', '.join(args)})")
    
    BUILD_TUPLE = BUILD_TUPLE_UNPACK = BUILD_TUPLE_UNPACK_WITH_CALL = build_tuple

    def build_list(self, inst: Instruction):
        args = [self.state.stack.pop() for _ in range(inst.argval)]
        args = args[::-1]
        if "UNPACK" in inst.opname:
            args = [f"*{arg}" for arg in args]
        self.state.stack.append(f"[{', '.join(args)}]")
        self.replace_mutable_tos_with_temp()
    
    BUILD_LIST = BUILD_LIST_UNPACK = build_list

    def build_set(self, inst: Instruction):
        ans = ""
        if inst.argval == 0:
            ans = "set()"
        else:
            args = [self.state.stack.pop() for _ in range(inst.argval)]
            args = args[::-1]
            if "UNPACK" in inst.opname:
                args = [f"*{arg}" for arg in args]
            ans = f"{{{', '.join(args)}}}"
        self.state.stack.append(ans)
        self.replace_mutable_tos_with_temp()
    
    BUILD_SET = BUILD_SET_UNPACK = build_set

    def build_map_unpack(self, inst: Instruction):
        if inst.argval == 0:
            self.state.stack.append("dict()")
        else:
            args = [self.state.stack.pop() for _ in range(inst.argval)]
            args = args[::-1]
            args = [f"**{arg}" for arg in args]
            self.state.stack.append(f"{{{', '.join(args)}}}")
        self.replace_mutable_tos_with_temp()

    BUILD_MAP_UNPACK = BUILD_MAP_UNPACK_WITH_CALL = build_map_unpack

    def BUILD_MAP(self, inst: Instruction):
        args = [self.state.stack.pop() for _ in range(inst.argval * 2)]
        args = args[::-1]
        keys = args[::2]
        values = args[1::2]
        self.state.stack.append(f"{{{', '.join([f'{k}: {v}' for k, v in zip(keys, values)])}}}")
        self.replace_mutable_tos_with_temp()

    def BUILD_CONST_KEY_MAP(self, inst: Instruction):
        keys = eval(self.state.stack.pop())
        args = [self.state.stack.pop() for _ in range(inst.argval)]
        values = args[::-1]
        self.state.stack.append(f"{{{', '.join([f'{k}: {v}' for k, v in zip(keys, values)])}}}")
        self.replace_mutable_tos_with_temp()

    def BUILD_STRING(self, inst: Instruction):
        args = [self.state.stack.pop() for _ in range(inst.argval)]
        args = args[::-1]
        values = " + ".join(args)
        self.state.stack.append(values)

    def LIST_TO_TUPLE(self, inst: Instruction):
        item = self.state.stack.pop()
        self.state.stack.append(f"tuple({item})")

    def LIST_EXTEND(self, inst: Instruction):
        assert inst.argval == 1, "Only tested for argval==1"
        values = self.state.stack.pop()
        temp = self.replace_mutable_tos_with_temp()
        self.state.source_code += f"{temp}.extend({values})\n"

    def LIST_APPEND(self, inst: Instruction):
        if inst.argval == 1:
            # it should be a bug, the tos should be the value. fix it anyway.
            inst.argval += 1
        container = self.state.stack[-inst.argval]
        value = self.state.stack.pop()
        self.state.source_code += f"{container}.append({value})\n"

    def generic_update(self, inst: Instruction):
        assert inst.argval == 1, "Only tested for argval==1"
        values = self.state.stack.pop()
        temp = self.replace_mutable_tos_with_temp()
        self.state.source_code += f"{temp}.update({values})\n"
    
    SET_UPDATE = DICT_UPDATE = DICT_MERGE = generic_update

    def SET_ADD(self, inst: Instruction):
        if inst.argval == 1:
            # it should be a bug, the tos should be the value. fix it anyway.
            inst.argval += 1
        container = self.state.stack[-inst.argval]
        value = self.state.stack.pop()
        self.state.source_code += f"{container}.add({value})\n"

    def MAP_ADD(self, inst: Instruction):
        container = self.state.stack[-inst.argval - 1]
        # see https://docs.python.org/3.10/library/dis.html#opcode-MAP_ADD
        if sys.version_info >= (3, 8):
            value = self.state.stack.pop()
            key = self.state.stack.pop()
        else:
            key = self.state.stack.pop()
            value = self.state.stack.pop()
        self.state.source_code += f"{container}.__setitem__({key}, {value})\n"

# ==================== Misc Instructions =============================
    def RAISE_VARARGS(self, inst: Instruction):
        if inst.argval == 0:
            self.state.source_code += "raise\n"
        elif inst.argval == 1:
            self.state.source_code += f"raise {self.state.stack.pop()}\n"
        elif inst.argval == 2:
            tos = self.state.stack.pop()
            tos1 = self.state.stack.pop()
            self.state.source_code += f"raise {tos1} from {tos}\n"
    
    def FORMAT_VALUE(self, inst: Instruction):
        func, spec = inst.argval
        if spec:
            form_spec = self.state.stack.pop()
            value = self.state.stack.pop()
            self.state.stack.append(f"format({value}, {form_spec})")
        else:
            value = self.state.stack.pop()
            func = str if func is None else func
            self.state.stack.append(f"{func.__name__}({value})")


# ==================== NOP Instructions =============================
    def generic_nop(self, inst: Instruction):
        pass

    # "EXTENDED_ARG" is treated as NOP here, because it has been handled by `dis.get_instructions`.
    # The extended args are already merged into the following instruction's `inst.argval`.
    EXTENDED_ARG = generic_nop

    NOP = RESUME = SETUP_LOOP = POP_BLOCK = PRECALL = BEGIN_FINALLY = END_FINALLY = generic_nop

    MAKE_CELL = generic_nop

    # our FOR_ITER is different from CPython's FOR_ITER (as it does not need to explicitly consider the case of exhausted iterator), so we don't need to do anything here
    END_FOR = generic_nop

# ==================== Unsupported Instructions =============================
    def unimplemented_instruction(self, inst: Instruction):
        raise NotImplementedError(f"Unsupported instruction: {inst.opname}")

    GET_YIELD_FROM_ITER = unimplemented_instruction

    # we don't support try-except/try-finally
    POP_EXCEPT = RERAISE = WITH_EXCEPT_START = JUMP_IF_NOT_EXC_MATCH = CHECK_EG_MATCH = PUSH_EXC_INFO = PREP_RERAISE_STAR = WITH_CLEANUP_FINISH = CALL_FINALLY = POP_FINALLY = WITH_CLEANUP_START = SETUP_EXCEPT = CHECK_EXC_MATCH = CLEANUP_THROW = unimplemented_instruction

    # we don't support async/await
    GET_AWAITABLE = GET_AITER = GET_ANEXT = END_ASYNC_FOR = BEFORE_ASYNC_WITH = SETUP_ASYNC_WITH = SEND = ASYNC_GEN_WRAP = unimplemented_instruction

    CACHE = unimplemented_instruction
        
    # we don't know these instructions
    PRINT_EXPR = COPY_DICT_WITHOUT_KEYS = unimplemented_instruction

    # we only support bytecode for functions
    IMPORT_STAR = unimplemented_instruction
    
    YIELD_FROM = SETUP_ANNOTATIONS = LOAD_BUILD_CLASS = SETUP_WITH = BEFORE_WITH = MATCH_MAPPING = MATCH_SEQUENCE = MATCH_KEYS = MATCH_CLASS = unimplemented_instruction

    # don't find any interesting use case for these instructions
    CALL_INTRINSIC_2 = unimplemented_instruction

    def decompile_range(self, start: int, end: int):
        running_index = start
        while running_index < end:
            inst = self.instructions[running_index]
            method = getattr(Decompiler, inst.opname, Decompiler.unimplemented_instruction)
            output = method(self, inst)
            if output:
                running_index = output
            else:
                running_index += 1

    def index_of(self, offset: int):
        for idx, inst in enumerate(self.instructions):
            if inst.offset == offset:
                return idx
        raise ValueError(f"Cannot find instruction with offset {offset}")

    @staticmethod
    def cleanup_instructions(instructions: List[Instruction]):
        propagate_line_nums(instructions)
        simplify_with_statement(instructions)
        simplify_finally_statement(instructions)
        nop_unreachable_bytecode(instructions)

    def __init__(self, code: Union[CodeType, Callable]):
        if callable(code):
            code = code.__code__
        self.code = code
        instructions = list(convert_instruction(_) for _ in dis.get_instructions(code))
        Decompiler.cleanup_instructions(instructions)
        self.instructions = instructions
        self.state = DecompilerState(source_code="", stack=[])

    def get_temp_name(self):
        Decompiler.temp_count += 1
        return f"{self.temp_prefix}{Decompiler.temp_count}"

    def replace_mutable_tos_with_temp(self):
        ans = self.state.stack.pop()
        temp_name = self.get_temp_name()
        self.state.source_code += f"{temp_name} = {ans}\n"
        self.state.stack.append(temp_name)
        return temp_name

    @staticmethod
    def supported_opnames():
        opnames = []
        for x in dis.opname:
            if getattr(Decompiler, x, Decompiler.unimplemented_instruction) is not Decompiler.unimplemented_instruction:
                opnames.append(x)
        return opnames

    @functools.lru_cache(maxsize=None)
    def decompile(self, indentation=4, temp_prefix: str="__temp_", overwite_fn_name: Optional[str]=None) -> str:
        self.indentation = indentation
        self.temp_prefix = temp_prefix
        self.decompile_range(0, len(self.instructions))
        source_code = self.state.source_code
        # the header might have invalid function name in torchdynamo. only optimize the function body.
        source_code = remove_some_temp(source_code, self.temp_prefix, indentation)
        header = get_function_signature(self.code, overwite_fn_name)
        source_code = header + add_indentation(source_code, indentation)
        return source_code

    def __hash__(self):
        return hash(self.code)



from typing import List, Tuple, Dict, Union, Callable, Optional
from types import CodeType

from .decompiler import Decompiler

def decompile(code: Union[CodeType, Callable]):
    """Decompile a code object or a function."""
    return Decompiler(code).decompile()


import types


def pytorch_bytecode_src_hook(code: types.CodeType, new_code: types.CodeType):
    import torch
    bytecode_log = torch._logging.getArtifactLogger(
        "torch._dynamo.convert_frame", "bytecode"
    )
    import logging

    if bytecode_log.isEnabledFor(logging.DEBUG):
        try:
            decompiled_src = decompile(new_code)
            bytecode_log.debug("possible source code:")
            bytecode_log.debug(decompiled_src)
        except Exception as e:
            bytecode_log.debug("Decompilation fails due to: %s", str(e))
        finally:
            bytecode_log.debug(
                "If you find the decompiled code is wrong,"
                "please submit an issue at "
                "https://github.com/youkaichao/depyf/issues."
            )

_handle = None

def install():
    import torch
    global _handle
    if _handle is not None:
        return
    _handle = torch._dynamo.convert_frame.register_bytecode_hook(pytorch_bytecode_src_hook)


def uninstall():
    global _handle
    if _handle is None:
        return
    _handle.remove()
    _handle = None

import os

__version__ = open(f"{os.path.dirname(__file__)}/VERSION.txt").read().strip()



import dis
from typing import List, Tuple, Union, Optional, Callable, Any, Dict, Set
from types import CodeType

def escape_html(s: str) -> str:
    """Escape string for use in HTML."""
    return (s.replace("&", "&amp;")
            .replace(" ", "&nbsp;")
            .replace("<", "&lt;")
            .replace(">", "&gt;")
            .replace('"', "&quot;")
            .replace("'", "&apos;"))

def generate_dot_table(header: str, rows: List[List[str]]) -> str:
    """
    Generate an HTML table string with a header spanning n columns and given rows.

    Args:
    - header (str): The header string.
    - rows (list of list of str): A list of rows, where each row is a list of n strings.

    Returns:
    - str: The generated HTML table string.
    """

    # Start the table
    html_str = '<TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">'

    n = len(rows[0])
    # Add the header spanning n columns
    html_str += f'<TR><TD  ALIGN="LEFT">{header}</TD>' + '<TD  ALIGN="LEFT"></TD>' * (n - 1) + '</TR>'

    # Add each row
    for row in rows:
        if len(row) != n:
            raise ValueError("Inconsistent number of columns.")
        html_str += '<TR>' + ''.join([f'<TD  ALIGN="LEFT">{escape_html(cell)}</TD>' for cell in row]) + '</TR>'

    # Close the table
    html_str += '</TABLE>'
    
    return "<\n" + html_str + "\n>"

def get_function_signature(code_obj: CodeType, overwite_fn_name: Optional[str]=None) -> str:
    # Extract all required details from the code object
    # Sometimes the code object does not have a name, e.g. when it is a lambda function, so we can overwrite it to be a valid name
    normal_arg_count = code_obj.co_argcount + code_obj.co_kwonlyargcount
    arg_names = code_obj.co_varnames[:normal_arg_count]
    arg_names = [x if not x.startswith(".") else x.replace(".", "comp_arg_") for x in arg_names]
    
    import inspect
    if code_obj.co_flags & inspect.CO_VARARGS:
        arg_names.append('*' + code_obj.co_varnames[normal_arg_count])
        normal_arg_count += 1
    if code_obj.co_flags & inspect.CO_VARKEYWORDS:
        arg_names.append('**' + code_obj.co_varnames[normal_arg_count])
        normal_arg_count += 1
    args_str = ', '.join(arg_names)
    fn_name = overwite_fn_name if overwite_fn_name is not None else code_obj.co_name
    header = f"def {fn_name}({args_str}):\n"
    return header



from setuptools import setup, find_packages

setup(
    name='depyf',
    version=open("depyf/VERSION.txt").read().strip(),
    description='Decompile python functions, from bytecode to source code!',
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url='https://github.com/youkaichao/depyf',
    author='Kaichao You',
    author_email="youkaichao@gmail.com",
    license="MIT",
    packages=find_packages(include=["depyf"]),
    python_requires=">=3.7",
    install_requires=[
        "astor",
    ],
    extras_require={
        "dev": [
            "pytest",
        ]
    }
)


import dis
from typing import List, Tuple, Union, Optional, Callable, Any, Dict, Set
from types import CodeType
import ast
import astor
from collections import defaultdict
import dataclasses
import sys

py311 = sys.version_info >= (3, 11)
all_jump_opcode_set = set(dis.hasjabs) | set(dis.hasjrel)


@dataclasses.dataclass
class Instruction:
    """A mutable version of dis.Instruction"""

    opcode: int
    opname: str
    arg: Optional[int]
    argval: Any
    argrepr: str
    offset: Optional[int] = None
    starts_line: Optional[int] = None
    is_jump_target: bool = False

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return id(self) == id(other)

    def short_inst_repr(self):
        return f"Instruction(opname={self.opname}, offset={self.offset})"

    def is_jump(self):
        return self.opcode in all_jump_opcode_set

    def get_jump_target(self: "Instruction"):
        if self.is_jump() and "to " in self.argrepr:
            return int(self.argrepr.replace("to ", "").strip())
        # seems like a bug, "FOR_ITER" is in `dis.hasjrel`, but its `argval` is an absolute offset
        if self.opcode in dis.hasjabs:
            return self.argval
        elif self.opcode in dis.hasjrel:
            return self.offset + self.argval if not py311 else self.argval
        else:
            raise ValueError(f"Instruction {self.opname} does not have jump target")


def convert_instruction(i: dis.Instruction) -> Instruction:
    return Instruction(
        i.opcode,
        i.opname,
        i.arg,
        i.argval,
        i.argrepr,
        i.offset,
        i.starts_line,
        i.is_jump_target,
    )


def nop_instruction(inst: Instruction):
    """Inplace modify an instruction as nop."""
    inst.opname = "NOP"
    inst.opcode = dis.opmap["NOP"]
    inst.arg = 0
    inst.argval = 0
    inst.argrepr = ""
    inst.offset
    inst.starts_line
    inst.is_jump_target = False
    return inst


def propagate_line_nums(instructions: List[Instruction]):
    """Ensure every instruction has line number set in case some are removed"""
    cur_line_no = None

    def populate_line_num(inst):
        nonlocal cur_line_no
        if inst.starts_line:
            cur_line_no = inst.starts_line

        inst.starts_line = cur_line_no

    for inst in instructions:
        populate_line_num(inst)


def simplify_with_statement(instructions: List[Instruction]):
    """Simplify with statement.
    3.10 with statement:
    SETUP_WITH
    with body
    POP_BLOCK
    some extra code (starts_line == with)
    """
    for i, inst in enumerate(instructions):
        if inst.opname == "SETUP_WITH":
            line_no = inst.starts_line
            pop_blocks = [j for j, _inst in enumerate(instructions) if j > i and _inst.opname == "POP_BLOCK" and instructions[j + 1].starts_line == line_no]
            if pop_blocks:
                pop_block_index = pop_blocks[0]
                nop_instruction(instructions[pop_block_index])
                for _inst in instructions[pop_block_index:]:
                    if _inst.starts_line == line_no:
                        nop_instruction(_inst)


def simplify_finally_statement(instructions: List[Instruction]):
    """Simplify finally statement.
    3.10 finally statement:
    SETUP_FINALLY
    body
    POP_BLOCK
    finally code
    Exception code
    RERAISE
    """
    for i, inst in enumerate(instructions):
        if inst.opname == "SETUP_FINALLY":
            finally_target = inst.get_jump_target()
            reraise_idx = [j for j, _inst in enumerate(instructions) if _inst.offset >= finally_target and _inst.opname == "RERAISE"]
            if reraise_idx:
                reraise_index = reraise_idx[0]
                for j, _inst in enumerate(instructions):
                    if _inst.offset >= finally_target and j <= reraise_index:
                        nop_instruction(_inst)


def nop_unreachable_bytecode(instructions: List[dis.Instruction]) -> List[dis.Instruction]:
    """Mark unreachable bytecode as NOP."""
    jumps = set(dis.hasjabs) | set(dis.hasjrel)

    reachable = [False for x in instructions]
    reachable[0] = True
    # each instruction marks the instruction after it
    for i, inst in enumerate(instructions):
        if inst.is_jump_target:
            # the instruction is the target of a jump
            reachable[i] = True
        # the last instruction does not need to mark any following instructions
        if i == len(instructions) - 1:
            break
        # this instruction is not reachable, nothing to do
        if not reachable[i]:
            continue
        # this instruction is reachable
        # the following instruction is reachable if it is sequential op or conditional jump
        if inst.opname in ["RETURN_VALUE", "BREAK_LOOP"]:
            # the instruction after the return is unreachable
            pass
        elif inst.opcode in jumps:
            if inst.opcode in dis.hasjrel and inst.get_jump_target() == inst.offset:
                # this is a jump to itself, it is regarded as a NOP, per the documentation at
                # https://devguide.python.org/internals/interpreter/#jumps
                reachable[i] = False
                reachable[i + 1] = True
                continue
            if "IF" in inst.opname or "FOR_ITER" in inst.opname or "SETUP_LOOP" in inst.opname:
                # the fallback block is always reachable for conditional jumps
                reachable[i + 1] = True
            elif inst.opname == "SETUP_FINALLY":
                # the finally block is always reachable
                reachable[i + 1] = True
            else:
                # this is a direct jump, the target is reachable
                # we further check if any other in-between instructions are jump targets
                # if not, we can mark this instruction as unreachable, too
                # later, in-between instructions will be marked as unreachable (NOP)
                # and the interpreter will slide through all the NOP directly to the target
                jump_forwards = [j for j, instruct in enumerate(instructions) if instruct.offset >= inst.get_jump_target()]
                if len(jump_forwards):
                    j = jump_forwards[0]
                    if j > i and all(not instruct.is_jump_target for instruct in instructions[i + 1:j]):
                        reachable[i] = False
        else:
            reachable[i + 1] = True
    
    # mark unreachable instructions as NOP
    for inst, flag in zip(instructions, reachable):
        if not flag:
            nop_instruction(inst)

def add_indentation(code: str, indentation: int = 4) -> str:
    """Add indentation to code."""
    return "".join(" " * indentation + line + "\n" for line in code.splitlines())

def remove_indentation(code: str, indentation: int = 4) -> str:
    """Remove indentation from code."""
    return "".join(line[indentation:] + "\n" for line in code.splitlines())


class RemoveAssignmentTransformer(ast.NodeTransformer):
    def __init__(self, temp_name: str, temp_occurrences: Dict[str, List[ast.Name]]):
        # optimize one temp_name at a time
        self.temp_name = temp_name
        self.temp_occurrences = temp_occurrences
    def visit_Assign(self, node):
        # single assimngment like `temp = xxx`
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            name = node.targets[0].id
            # the assignment is like `temp = xxx`
            if name == self.temp_name:
                if len(self.temp_occurrences[name]) == 1:
                    return ast.Expr(value=node.value)
                elif len(self.temp_occurrences[name]) == 3 and isinstance(self.temp_occurrences[name][-1], bool):
                    # we save the `xxx` here
                    self.temp_occurrences[name].append(node.value)
                    if self.temp_occurrences[name][-2]:
                        return None
        return node

class RemoveAssignment2Transformer(ast.NodeTransformer):
    def __init__(self, temp_name: str, temp_occurrences: Dict[str, List[ast.Name]]):
        # optimize one temp_name at a time
        self.temp_name = temp_name
        self.temp_occurrences = temp_occurrences
    def visit_Name(self, node):
        name = node.id
        if name == self.temp_name and len(self.temp_occurrences[name]) == 4 and isinstance(self.temp_occurrences[name][-2], bool):
            if self.temp_occurrences[name][-2]:
                return self.temp_occurrences[name][-1]
        return node

def get_parents(node):
    """Collect all parent nodes of a given node."""
    parents = []
    while node:
        parents.append(node)
        node = getattr(node, "parent", None)
    return parents

def set_parents(node, parent=None):
    """Recursively set the parent attribute for each node."""
    for child in ast.iter_child_nodes(node):
        child.parent = parent
        set_parents(child, child)

def lowest_common_parent(node1, node2):
    """Get the lowest common parent for two nodes."""
    parents1 = get_parents(node1)
    parents2 = get_parents(node2)
    
    # Reverse the parents list to start comparing from the root.
    parents1.reverse()
    parents2.reverse()
    
    last_common = None
    for p1, p2 in zip(parents1, parents2):
        if p1 is p2:
            last_common = p1
        else:
            break
    return last_common, p1, p2

def remove_some_temp(source_code: str, temp_prefix:str, indentation: int=4) -> str:
    tree = ast.parse(source_code)
    set_parents(tree)

    temp_occurrences = defaultdict(list)
    for node in ast.walk(tree):
        if isinstance(node, ast.Name) and node.id.startswith(temp_prefix):
            temp_occurrences[node.id].append(node)

    for key in temp_occurrences:
        if len(temp_occurrences[key]) == 2:
            node1 = temp_occurrences[key][0]
            node2 = temp_occurrences[key][1]
            parent, parent1, parent2 = lowest_common_parent(node1, node2)
            assignment_node = node1 if isinstance(node1.parent, ast.Assign) else node2
            assignment_parent = parent1 if isinstance(node1.parent, ast.Assign) else parent2
            indentation_nodes = (ast.FunctionDef, ast.AsyncFunctionDef, ast.For, ast.AsyncFor, ast.While, ast.If, ast.Try, ast.With, ast.AsyncWith, ast.ClassDef)
            # we cannot remove the assignment if the assignment `temp=xxx` is in an indentation block while the usage of `temp` is not
            can_merge = not isinstance(assignment_parent, indentation_nodes)
            temp_occurrences[key].append(can_merge)
        tree = RemoveAssignmentTransformer(key, temp_occurrences).visit(tree)
        tree = RemoveAssignment2Transformer(key, temp_occurrences).visit(tree)

    reconstructed_code = astor.to_source(tree, indent_with=" " * indentation)
    return reconstructed_code
